/*********************************************************************
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2008, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Willow Garage nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************/

#pragma once

/***************************************************/
/*! \class controller::FrequencyResponseController
    \brief Frequency Response Controller

    This tests the moment of inertia and the damping of a 
    joint using a sinusoidal torque input

*/
/***************************************************/


#include <ros/node.h>
#include <math.h>
#include <dynamic_verification_controllers/DynamicResponseData.h>
#include <realtime_tools/realtime_publisher.h>
#include <realtime_tools/realtime_srv_call.h>
#include <mechanism_model/controller.h>
#include <robot_mechanism_controllers/joint_velocity_controller.h>

namespace controller
{

class FrequencyResponseController : public Controller
{

public:
  FrequencyResponseController();
  ~FrequencyResponseController();

  /*!
   * \brief Functional way to initialize.
   * \param velocity Target velocity for the velocity controller.
   * \param max_effort Effort to limit the controller at.
   * \param *robot The robot that is being controlled.
   */
  void init( double effort, double test_duration, double frequency, double settle_time, double time, std::string name, mechanism::RobotState *robot);
  bool initXml(mechanism::RobotState *robot, TiXmlElement *config);


  /*!
   * \brief Perform the test analysis
   */
  void analysis();

  /*!
   * \brief Issues commands to the joint. Should be called at regular intervals
   */
  virtual void update();
  
  bool done() { return done_; }
  
  dynamic_verification_controllers::DynamicResponseData::Request dynamic_data_;

private:

  mechanism::JointState *joint_state_;      /**< Joint we're controlling. */
  mechanism::RobotState *robot_;            /**< Pointer to robot structure. */
  double effort_;                  /**< Torque applied during the test. */
  double frequency_;              /**< Frequency of sinesweep. */

  double initial_time_;                     /**< Start time of the test. */
  double initial_position_;
  int count_;
  bool complete;
  bool start;
  double test_duration_;
  double settle_time_;

  int done_;
  int starting_count;

};

/***************************************************/
/*! \class controller::FrequencyResponseControllerNode
    \brief Frequency Response Controller

    This tests the dynamic response of a joint using 
    a sinusoidal input. Applies a sinusoidal response, 
    waits for system to settle before measuring. Frequency in Hz. 

<controller type="FrequencyResponseControllerNode" name="r_wrist_flex_freq_controller">
  <joint name="r_wrist_flex_joint">
    <controller_defaults effort="0.1" frequency="20.0" duration="10" settle_time="10" />
  </joint>
</controller>

*/
/***************************************************/

class FrequencyResponseControllerNode : public Controller
{
public:

  FrequencyResponseControllerNode();
  ~FrequencyResponseControllerNode();
   
  void update();
  bool initXml(mechanism::RobotState *robot, TiXmlElement *config);

private:
  FrequencyResponseController *c_;
  mechanism::RobotState *robot_;
  
  bool data_sent_;
  
  realtime_tools::RealtimeSrvCall<dynamic_verification_controllers::DynamicResponseData::Request, dynamic_verification_controllers::DynamicResponseData::Response> call_service_;

};
}


