/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "planning_display.h"
#include "common.h"
#include "robot/robot.h"
#include "properties/property.h"
#include "properties/property_manager.h"

#include <ogre_tools/axes.h>

#include <tf/transform_listener.h>
#include <planning_environment/robot_models.h>

#include <OGRE/OgreSceneNode.h>
#include <OGRE/OgreSceneManager.h>

namespace rviz
{

PlanningDisplay::PlanningDisplay( const std::string& name, VisualizationManager* manager )
: Display( name, manager )
, env_models_( NULL )
, kinematic_model_( NULL )
, new_kinematic_path_( false )
, animating_path_( false )
, state_display_time_( 0.05f )
{
  robot_ = new Robot( vis_manager_ );

  setVisualVisible( false );
  setCollisionVisible( true );
  robot_->setUserData( Ogre::Any( (void*)this ) );
}

PlanningDisplay::~PlanningDisplay()
{
  unsubscribe();

  delete robot_;
}

void PlanningDisplay::initialize( const std::string& description_param, const std::string& kinematic_path_topic )
{
  setRobotDescription( description_param );
  setTopic( kinematic_path_topic );
}

void PlanningDisplay::setRobotDescription( const std::string& description_param )
{
  description_param_ = description_param;

  propertyChanged(robot_description_property_);

  if ( isEnabled() )
  {
    load();
    causeRender();
  }
}

void PlanningDisplay::setTopic( const std::string& topic )
{
  unsubscribe();
  kinematic_path_topic_ = topic;
  subscribe();

  propertyChanged(topic_property_);
}

void PlanningDisplay::setStateDisplayTime( float time )
{
  state_display_time_ = time;

  propertyChanged(state_display_time_property_);

  causeRender();
}

void PlanningDisplay::setVisualVisible( bool visible )
{
  robot_->setVisualVisible( visible );

  propertyChanged(visual_enabled_property_);

  causeRender();
}

void PlanningDisplay::setCollisionVisible( bool visible )
{
  robot_->setCollisionVisible( visible );

  propertyChanged(collision_enabled_property_);

  causeRender();
}

bool PlanningDisplay::isVisualVisible()
{
  return robot_->isVisualVisible();
}

bool PlanningDisplay::isCollisionVisible()
{
  return robot_->isCollisionVisible();
}

void PlanningDisplay::load()
{
  std::string content;
  ros_node_->getParam(description_param_, content);

  TiXmlDocument doc;
  doc.Parse(content.c_str());
  if (!doc.RootElement())
  {
    ROS_ERROR("Failed to load robot model from description: %s", description_param_.c_str());
    return;
  }

  mechanism::Robot descr;
  descr.initXml(doc.RootElement());
  robot_->load( descr );
  
  delete env_models_;
  
  env_models_ = new planning_environment::RobotModels(description_param_);
  kinematic_model_ = env_models_->getKinematicModel().get();
  
  robot_->update( kinematic_model_, target_frame_ );
}

void PlanningDisplay::onEnable()
{
  subscribe();

  load();
  robot_->setVisible( true );
}

void PlanningDisplay::onDisable()
{
  unsubscribe();
  robot_->setVisible( false );
}

void PlanningDisplay::subscribe()
{
  if ( !isEnabled() )
  {
    return;
  }

  if ( !kinematic_path_topic_.empty() )
  {
    ros_node_->subscribe( kinematic_path_topic_, incoming_kinematic_path_message_, &PlanningDisplay::incomingKinematicPath, this, 0 );
  }

}

void PlanningDisplay::unsubscribe()
{
  if ( !kinematic_path_topic_.empty() )
  {
    ros_node_->unsubscribe( kinematic_path_topic_, &PlanningDisplay::incomingKinematicPath, this );
  }
}

void PlanningDisplay::update(float wall_dt, float ros_dt)
{
  if(!kinematic_model_)
    return;

  incoming_kinematic_path_message_.lock();

  if ( !animating_path_ && new_kinematic_path_ )
  {
    displaying_kinematic_path_message_ = incoming_kinematic_path_message_;

    animating_path_ = true;
    new_kinematic_path_ = false;
    current_state_ = -1;
    current_state_time_ = state_display_time_ + 1.0f;

    if (displaying_kinematic_path_message_.start_state.vals.size() > 0)
	kinematic_model_->computeTransforms(&displaying_kinematic_path_message_.start_state.vals[0]);
    robot_->update( kinematic_model_, target_frame_ );
  }

  incoming_kinematic_path_message_.unlock();

  if ( animating_path_ )
  {
    if ( current_state_time_ > state_display_time_ )
    {
      ++current_state_;

      calculateRobotPosition();

      if ( (size_t)current_state_ < displaying_kinematic_path_message_.get_states_size() )
      {
        int group_id = kinematic_model_->getGroupID( displaying_kinematic_path_message_.model_id );
	if (displaying_kinematic_path_message_.states[ current_state_ ].vals.size() > 0)
	    kinematic_model_->computeTransformsGroup(&displaying_kinematic_path_message_.states[ current_state_ ].vals[0], group_id);
        robot_->update( kinematic_model_, target_frame_ );

        causeRender();
      }
      else
      {
        animating_path_ = false;
      }

      current_state_time_ = 0.0f;
    }

    current_state_time_ += wall_dt;
  }
}

void PlanningDisplay::calculateRobotPosition()
{
  tf::Stamped<tf::Pose> pose( btTransform( btQuaternion( 0, 0, 0 ), btVector3( 0, 0, 0 ) ), 
			      displaying_kinematic_path_message_.header.stamp, displaying_kinematic_path_message_.header.frame_id );

  if (tf_->canTransform(target_frame_, displaying_kinematic_path_message_.header.frame_id, displaying_kinematic_path_message_.header.stamp))
  {
    try
    {
      tf_->transformPose( target_frame_, pose, pose );
    }
    catch(tf::TransformException& e)
    {
      ROS_ERROR( "Error transforming from frame '%s' to frame '%s'", pose.frame_id_.c_str(), target_frame_.c_str() );
    }
  }

  Ogre::Vector3 position( pose.getOrigin().x(), pose.getOrigin().y(), pose.getOrigin().z() );
  robotToOgre( position );

  btScalar yaw, pitch, roll;
  pose.getBasis().getEulerZYX( yaw, pitch, roll );
  Ogre::Matrix3 orientation;
  orientation.FromEulerAnglesYXZ( Ogre::Radian( yaw ), Ogre::Radian( pitch ), Ogre::Radian( roll ) );

  robot_->setPosition( position );
  robot_->setOrientation( orientation );
}

void PlanningDisplay::incomingKinematicPath()
{
  new_kinematic_path_ = true;
}

void PlanningDisplay::targetFrameChanged()
{
  calculateRobotPosition();
}

void PlanningDisplay::createProperties()
{
  visual_enabled_property_ = property_manager_->createProperty<BoolProperty>( "Visual Enabled", property_prefix_, boost::bind( &PlanningDisplay::isVisualVisible, this ),
                                                                               boost::bind( &PlanningDisplay::setVisualVisible, this, _1 ), category_, this );
  collision_enabled_property_ = property_manager_->createProperty<BoolProperty>( "Collision Enabled", property_prefix_, boost::bind( &PlanningDisplay::isCollisionVisible, this ),
                                                                                 boost::bind( &PlanningDisplay::setCollisionVisible, this, _1 ), category_, this );
  state_display_time_property_ = property_manager_->createProperty<FloatProperty>( "State Display Time", property_prefix_, boost::bind( &PlanningDisplay::getStateDisplayTime, this ),
                                                                                  boost::bind( &PlanningDisplay::setStateDisplayTime, this, _1 ), category_, this );
  FloatPropertyPtr float_prop = state_display_time_property_.lock();
  float_prop->setMin( 0.0001 );

  robot_description_property_ = property_manager_->createProperty<StringProperty>( "Robot Description", property_prefix_, boost::bind( &PlanningDisplay::getRobotDescription, this ),
                                                                                   boost::bind( &PlanningDisplay::setRobotDescription, this, _1 ), category_, this );
  topic_property_ = property_manager_->createProperty<ROSTopicStringProperty>( "Topic", property_prefix_, boost::bind( &PlanningDisplay::getTopic, this ),
                                                                               boost::bind( &PlanningDisplay::setTopic, this, _1 ), category_, this );
  ROSTopicStringPropertyPtr topic_prop = topic_property_.lock();
  topic_prop->setMessageType(motion_planning_msgs::KinematicPath::__s_getDataType());

  robot_->setPropertyManager( property_manager_, category_ );
}

const char* PlanningDisplay::getDescription()
{
  return "Displays a planned path given by a motion_planning_msgs::KinematicPath message.";
}

} // namespace rviz


